import pyfcm
from pyfcm import FCMNotification


#https://github.com/olucurious/PyFCM
def send_notification(registration_ids, message_body):

    push_service = FCMNotification(api_key="AAAAQEUZt1M:APA91bHctxd-8I9Mdc2bdIfo5OYm9leUfPOnVFzVg2xQ7xFwTPbsx5HwmzGx6TiHZzWBKaii8CRJ6NKhimnSR8X3Z0g6PEpsutpQekBm2-XbMKSDKcfHz7-3PhjFXHz7GM4e0CrbMAjc")
    result = push_service.multiple_devices_data_message(registration_ids=registration_ids, data_message=message_body)
    return result


#!/usr/bin/python
# -*- coding: utf-8 -*-
import base64
import hashlib
import json
import random
import traceback
import urllib.request
import uuid
import os

BUFFER_SIZE = 64

END_OF_MESSAGE = '[END_OF_MESSAGE]'


def is_server_premium(remote_id, access_token):
    try:
        url = "http://cloud.homevee.de/server-api.php?action=ispremium&remoteid=" + remote_id + "&accesstoken=" + access_token
        contents = urllib.request.urlopen(url).read().decode('utf-8')

        if contents == "wrongcredentials":
            return None

        data = json.loads(contents)

        if 'is_premium' in data and data['is_premium'] is not None:
            return data['is_premium']
        else:
            return None
    except:
        traceback.print_exc()
        return None


def assign_cloud_to_remote_id(cloud, server_secret, remote_id, access_token):
    try:
        url = "http://cloud.homevee.de/server-api.php?action=assigncloudtoremoteid&remoteid=" + remote_id + "&cloud=" + cloud + "&secret=" + server_secret + "&accesstoken=" + access_token

        print(url)

        contents = urllib.request.urlopen(url).read().decode('utf-8')

        if (contents == "wrongcredentials"):
            return False, False

        print("response: " + contents)

        data = json.loads(contents)

        verified = data['verified']
        is_premium = data['is_premium']
        status = data['status']

        if status == "nopermission":
            return False, False

        return verified, is_premium

    except:
        traceback.print_exc()

    return False, False


def encrypt_data(data, private_key):
    private_key = ""
    return data


def decrypt_data(data, public_key):
    private_key = ""
    return data


def send_to_server(conn, text):
    len_send = conn.send(text + END_OF_MESSAGE)

    print
    "sending to local server: " + text

    result = ""

    while not result.endswith(END_OF_MESSAGE):
        result += conn.recv(BUFFER_SIZE)

    print
    "received from local server: " + result

    return result


def get_client_id(clients):
    i = 0
    while True:
        if i in clients:
            i += 1
        else:
            return i

def generate_server_credentials(db):
    remote_id = ""
    access_token = ""

    # generate random remote_id thats not in db
    with db:
        cur = db.cursor()

        lst = [random.choice('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ') for _ in range(8)]
        id = "".join(lst)

        cur.execute("SELECT REMOTE_ID FROM SERVER_DATA WHERE REMOTE_ID = :id", {'id': id})
        data = cur.fetchone()
        if data is None:
            # no entry existing
            remote_id = id

    # generate access_token
    random_bytes = os.urandom(128)
    access_token = base64.b64encode(random_bytes).decode('utf-8')

    print("Connected to: "+remote_id)

    # hash_password
    salt = base64.urlsafe_b64encode(uuid.uuid4().bytes)
    t_sha = hashlib.sha512()
    t_sha.update(access_token + salt)
    hashed_token = base64.urlsafe_b64encode(t_sha.digest())

    save_server_access_token(remote_id, hashed_token, salt, db)

    return remote_id, access_token

    '''
        len_send = conn.send(text)

        data = conn.recv(8192)

        return data
    '''


import argparse

import HomeveeCloud


def main():
    parser = argparse.ArgumentParser(description='Homevee-Cloud for Servers')
    parser.add_argument('--domain', required=True, type=str, help='Domain that the server is reachable under (z.B. free.cloud.homevee.de)')
    parser.add_argument('--port', default=7778, type=int, help='The Port the Homevee-Server should listen on')
    parser.add_argument('--server_secret', required=True, type=str, help='The secret for communication with the main server')
    parser.add_argument('--max_clients', default=50, type=int, help='Number of clients simultaneously')
    parser.add_argument('--buffer_size', default=64, type=int, help='Buffer size for client communication')
    parser.add_argument('--debug', default=False, type=bool, help='Is the server in debug mode?')
    args = parser.parse_args()

    #homevee_cloud = HomeveeCloud(args.server_secret, args.domain, args.port, args.max_clients, args.debug, args.buffer_size)
    #homevee_cloud.start_server()

    HomeveeCloud.DOMAIN = args.domain
    HomeveeCloud.PORT = args.port
    HomeveeCloud.CLIENT_NUM = args.max_clients
    HomeveeCloud.IS_DEBUG = args.debug
    HomeveeCloud.BUFFER_SIZE = args.buffer_size
    HomeveeCloud.SERVER_SECRET = args.server_secret

    HomeveeCloud.start()

if __name__ == "__main__":
    main()


'''import BaseHTTPServer, SimpleHTTPServer
import ssl

if __name__ == "__main__":
    ssl_key = '/etc/letsencrypt/live/cloud.homevee.de/privkey.pem'
    ssl_cert = '/etc/letsencrypt/live/cloud.homevee.de/fullchain.pem'

    httpd = BaseHTTPServer.HTTPServer(('localhost', 9999),SimpleHTTPServer.SimpleHTTPRequestHandler)
    #httpd.socket = ssl.wrap_socket (httpd.socket,keyfile=ssl_key,certfile=ssl_cert, server_side=True)
    httpd.serve_forever()
'''

import socket
import traceback
from socket import AF_INET, SOCK_STREAM, SO_REUSEADDR, SOL_SOCKET, SHUT_RDWR
import ssl

KEYFILE = '/etc/letsencrypt/live/cloud.homevee.de/privkey.pem'
CERTFILE = '/etc/letsencrypt/live/cloud.homevee.de/fullchain.pem'

def echo_client(s):
    while True:
        data = s.recv(8192)
        print(data.decode("utf-8"))
        if data == b'':
            break
        s.send(b'This is a response.')
        print('Sendind: This is a response.')
        print('Connection closed')
    s.close()

def echo_server(address):
    s = socket.socket(AF_INET, SOCK_STREAM)
    s.bind(address)
    s.listen(1)
    s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

    s_ssl = ssl.wrap_socket(s, keyfile=KEYFILE, certfile=CERTFILE, server_side=True)

    while True:
        try:
            (c,a) = s_ssl.accept()
            print('Got connection', c, a)
            echo_client(c)
        except socket.error as e:
            traceback.print_exc()
            print('Error: {0}'.format(e))

echo_server((socket.gethostbyname('cloud.homevee.de'), 7778))


import json
import socket
import ssl
import sys
import traceback
from io import StringIO
from _thread import start_new_thread
from http.server import BaseHTTPRequestHandler

from HomeveeCloud import socket_utils, firebase_utils

SOCKET_MAP = {}

SERVER_SECRET = None
DOMAIN = None
PORT = 7778
CLIENT_NUM = 50
BUFFER_SIZE = 64
IS_DEBUG = False

def start():
    print("Starting Homevee-Server")
    listening_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        domain = "localhost"

        if(not IS_DEBUG):
            domain = socket.gethostbyname(DOMAIN)
        listening_socket.bind((domain, PORT))
    except socket.error as msg:
        print(msg)
        print(sys.exit())

    listening_socket.listen(CLIENT_NUM)

    while True:
        print(listening_socket.fileno())

        try:
            listen_for_requests(listening_socket)
        except(ssl.SSLError):
            print("SSLError")
        except(socket.error):
            #traceback.print_exc()
            print("SocketError")
            break
        except KeyboardInterrupt:
            print("CTRL-C => Exiting")
            break
        except:
            traceback.print_exc()

    listening_socket.close()

def listen_for_requests(listening_socket):
    #if(not IS_DEBUG):
    KEYFILE = '/etc/letsencrypt/live/' + DOMAIN + '/privkey.pem'
    CERTFILE = '/etc/letsencrypt/live/' + DOMAIN + '/cert.pem'

    print(KEYFILE)
    print(CERTFILE)

    listening_socket = ssl.wrap_socket(listening_socket, keyfile=KEYFILE, certfile=CERTFILE, server_side=True)

    conn, addr = listening_socket.accept()
    print('Connected with ' + addr[0] + ':' + str(addr[1]))

    start_new_thread(clientthread, (conn,))

def clientthread(conn):
    # infinite loop so that function do not terminate and thread do not end.

    sent_data = None

    while True:

        is_http = False
        is_resend = False
        is_notification = False

        error = False

        try:
            # Receiving from client
            data = ""

            while(not data.endswith(socket_utils.END_OF_MESSAGE)):
                new_data =  conn.recv(BUFFER_SIZE)

                data = data + new_data

            if data == "":
                conn.close()
                return

            print("")

            print("Received: " + data)

            if data.endswith(socket_utils.END_OF_MESSAGE):
                data = data[:-(len(socket_utils.END_OF_MESSAGE))]

            try:
                DATA_DICT = json.loads(data)

                if 'remote_id' in DATA_DICT and 'access_token' in DATA_DICT:
                    remote_id = DATA_DICT['remote_id']

                    try:
                        print('server connected')
                        access_token = DATA_DICT['access_token']
                        #data = DATA_DICT['data']

                        verified, is_premium = socket_utils.assign_cloud_to_remote_id(DOMAIN, SERVER_SECRET, remote_id, access_token)

                        if verified:
                            print("id "+remote_id+" is verified")
                        else:
                            print("id "+remote_id+" is not verified")

                        if is_premium:
                            print("id "+remote_id+" is premium")
                        else:
                            print("id "+remote_id+" is not premium")

                        if verified:
                            #do stuff
                            SOCKET_MAP[remote_id] = {}
                            SOCKET_MAP[remote_id]['server'] = conn
                            SOCKET_MAP[remote_id]['clients'] = {}
                            send_to_client(json.dumps({'status': 'connectionok'}), conn, is_http)
                        else:
                            print("Not authorized")
                            result_dict = {'result': 'error'}
                            #len_send = conn.send(encrypt_data(json.dumps(result_dict)))
                            len_send = conn.send(json.dumps(result_dict))
                        return
                    except:
                        print("deleting remote id: "+remote_id)
                        del SOCKET_MAP[remote_id]

                elif 'remote_id' in DATA_DICT:
                    print('client connected')
                    try:
                        client_id = socket_utils.get_client_id(SOCKET_MAP[DATA_DICT['remote_id']]['clients'])
                        SOCKET_MAP[DATA_DICT['remote_id']]['clients'][client_id] = {}
                        SOCKET_MAP[DATA_DICT['remote_id']]['clients'][client_id]['type'] = 'socket'
                        SOCKET_MAP[DATA_DICT['remote_id']]['clients'][client_id]['connection'] = conn
                        print("processing socket data...")
                        process_data(client_id, DATA_DICT)
                    except(KeyError):
                        if 'remote_id' not in DATA_DICT:
                            #remote_id nich angegeben
                            traceback.print_exc()
                            data = {'result': 'noremoteid'}
                            send_to_client(json.dumps(data), conn, False)
                        elif DATA_DICT['remote_id'] not in SOCKET_MAP:
                            #local server not connected
                            traceback.print_exc()
                            data = {'result': 'servernotconnected'}
                            send_to_client(json.dumps(data), conn, False)
                    return
                elif 'registration_ids' in DATA_DICT:
                    is_notification = True
                    print('Push-Notification Command')
                    #Send push notifications
                    registration_ids = DATA_DICT['registration_ids']
                    message_data = json.loads(DATA_DICT['message_data'])

                    firebase_utils.send_notification(registration_ids, message_data)

                    send_to_client(json.dumps({'status': 'ok'}), conn, is_http)

                # Ist HTTP-Aufruf?
                try:
                    data = parse_http_headers(data)
                    is_http = True
                except AttributeError:
                    try:
                        data = json.loads(data)
                        print("Request is not in HTTP-Format")
                    except:
                        print("Data could not be parsed")
                        send_to_client(json.dumps({'status': 'error'}), conn, is_http)
                        error = True

                if 'resend' in data and data['resend'] is True:
                    # Daten erneut senden
                    print("Resend!")
                    is_resend = True
                    print(str(len(sent_data)) + " | " + sent_data)

                    send_to_client(sent_data[data['resendstart']:], conn, is_http)

                if not error and not is_resend and not is_notification:
                    if not is_resend and not is_notification:
                        reply = process_data(data)

                    if (reply is not None):
                        send_to_client(reply, conn, is_http)

                        sent_data = reply
                    else:
                        print("Error in reply")
                        send_to_client(json.dumps({'status': 'error'}), conn, is_http)

                    if not data:
                        break
            except:
                traceback.print_exc()
                if data == "registration":
                    #register server to cloud

                    #generate remote_id and access_token
                    #REMOTE_ID, ACCESS_TOKEN = socket_utils.generate_server_credentials(db)

                    #send_to_client(json.dumps({'result': 'registrationok', 'remote_id': REMOTE_ID, 'access_token': ACCESS_TOKEN}), conn, is_http)

                    send_to_client(json.dumps({'result': 'error', 'msg': 'registrationnotpossible'}), conn, is_http)

        except(socket.error):
            traceback.print_exc()
            break
        except:
            traceback.print_exc()
            send_to_client(json.dumps({'status': 'error'}), conn, is_http)

    # came out of loop
    conn.close()
    #db.commit()
    print("Verbindung beendet")

class HTTPRequest(BaseHTTPRequestHandler):
    def __init__(self, request_text):
        self.rfile = StringIO(request_text)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()

    def send_error(self, code, message):
        self.error_code = code
        self.error_message = message

def parse_http_headers(data):
    if data.startswith("GET"):
        print("HTTP-GET-Request")
        request = HTTPRequest(data).path
        path, param_string = request.split('?', 1)
    elif data.startswith("POST"):
        print("HTTP-POST-Request")
        lines = data.split('\n')
        param_string = lines[-1]
    else:
        raise AttributeError

    params = {}

    param_fragments = param_string.split('&')

    for param in param_fragments:
        if '=' in param:
            key, value = param.split('=')
            params[key] = value

    return params

def get_socket(remote_id):
    if remote_id in SOCKET_MAP:
        return SOCKET_MAP[remote_id]['server']
    else:
        return None

def process_data(client_id, data):
    data['client_id'] = client_id

    try:
        if 'remote_id' in data:
            remote_id = data['remote_id']
        else:
            return None

        conn = get_socket(remote_id)

        if conn is None:
            return {'status': 'error', 'msg': 'servernotconnected'}

        try:
            msg = socket_utils.send_to_server(conn, json.dumps(data))
        except:
            data = {'error': 'servernotconnected'}
            msg = json.dumps(data)
            send_to_client(msg, SOCKET_MAP[remote_id]['clients'][client_id]['connection'], False)
            del SOCKET_MAP[remote_id]['clients'][client_id]
            return

        print('Sent to client with id ' + str(client_id) + ': ' + msg)

        if SOCKET_MAP[remote_id]['clients'][client_id]['type'] == "websocket":
            print
            "websocket send"
            SOCKET_MAP[remote_id]['clients'][client_id]['server'].send_message_to_all(msg)
            # SOCKET_MAP[remote_id]['clients'][client_id]['server'].send_message(SOCKET_MAP[remote_id]['clients'][client_id]['connection'], msg)
        elif SOCKET_MAP[remote_id]['clients'][client_id]['type'] == "socket":
            print
            "socket send"
            send_to_client(msg, SOCKET_MAP[remote_id]['clients'][client_id]['connection'], False)
            SOCKET_MAP[remote_id]['clients'][client_id]['connection'].close()

        # delete client from socket_map
        del SOCKET_MAP[remote_id]['clients'][client_id]

        return
    except ValueError as e:
        traceback.print_exc()
        return

def send_to_client(data, conn, is_http):
    if not is_http:
        print("Sent Response: " + data)
    elif is_http:
        data = 'HTTP/1.1 200 OK\nContent-Type: text/html\n' + data + '\n'
        print("Sent HTTP-Response: " + data)

    # Prüfen, ob alle Daten gesendet wurden
    try:
        len_send = conn.send(data.encode('utf-8'))
        print("Data: " + str(len(data)) + " | Sent: " + str(len_send))
    except:
        traceback.print_exc()
        return

    # if(len_send is 0):
    #    send_to_client(json.dumps({'status': 'error'}), conn, is_http)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Homevee-Cloud",
    version="0.0.0.13",
    author="Homevee-Cloud",
    author_email="support@homevee.de",
    description="Dein neues Smarthome-System!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/homevee/homevee-cloud",
    entry_points = {
        #'console_scripts': ['HomeveeCloud=HomeveeCloud.command_line:main'],
        'console_scripts': ['HomeveeCloudLegacy=HomeveeCloudLegacy:main'],
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pyfcm',
    ]
)

#Commands to build
#sudo apt-get install python3-pip python3-dev libffi-dev libssl-dev libxml2-dev libxslt1-dev libjpeg8-dev zlib1g-dev -y
#pip install --upgrade setuptools wheel twine
